/* $Id: fileio.c,v 1.25 2001/07/26 17:31:57 jm Exp $
 * Configuration file I/O functions
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include <sys/types.h> /* required for netinet/in.h in AMIGAOS */
#include <netinet/in.h>
#include <arpa/inet.h>

#include <syslog.h>

#include "fileio.h"

/* defines */

#define ASSERT assert

#define BUFLEN 128                   /* input buffer length to start with */

#define DEBUGSAVE

#define IP_ADDR_LEN 16

/* Load data from an ASCII file
 *
 * Gets the keyword and data data from each line and calls a user-specified
 * function to process the line
 *
 * INPUT VALUES:
 *
 *    ) Userdata may be anything, just passed to the processing function
 *    ) File must not be NULL, internal error, ASSERT used
 * M01) Not enough memory for buffer
 * M02) Blank lines
 * M03) Comment in the beginning of the line
 * M04) Comment not in the beginning of the line
 * M05) Configuration entry before a comment
 * M06) The line is longer than the default buffer size
 * M07) Spaces before the keyword
 *
 */

int load_data(void *userdata, FILE *file,  int (*process) (void *, char *, char *))
{
	int i, result, len, flag, bufsize;
	char *buf, *key, *data, *temp;

	ASSERT(file != NULL);
	bufsize = BUFLEN;
	buf = malloc(bufsize);
	if (buf == NULL) {
		fprintf(stderr,"load_data: Not enough memory for buffer "
		"(%d bytes)!\n", bufsize);
		return FALSE;
	}

	do {
		if (fgets(buf, bufsize, file) == NULL) 
			break;  //  д´ 
		len = strlen(buf);
		ASSERT(len <= bufsize - 1);

		//line Bufũ⺸ ū,  ÷ д´
		while (buf[len - 1] != '\n') {
			bufsize *= 2;
			temp = realloc(buf, bufsize);
			if (temp == NULL) {
				fprintf(stderr, "load_data: Not enough memory for "
				"buffer (%d bytes)!\n", bufsize);
				free(buf);
				return FALSE;
			}
			buf = temp;
			if (fgets(buf + len, bufsize - len, file) == NULL) {
				break;
			}
			len = strlen(buf);
			ASSERT(len <= bufsize - 1);
			if (len < bufsize - 1 && buf[len - 1] != '\n') {
				fprintf(stderr, "load_data: Binary file "
				"containing string terminator or too long line\n");
				free(buf);
				return FALSE;
			}
		}

		if (buf[len - 1] != '\n') {
			fprintf(stderr,"load_data: Line is not terminated with a "
			"linefeed\n");
			free(buf);
			return FALSE;
		}
		buf[len - 1] = '\0'; /* remove trailing newline */
		data = NULL;
		flag = FALSE;

		//space ǳʶڴ.
		for (i = 0; i < strlen(buf); i++) {
			if (buf[i] != ' ') break;
		}
		key = buf + i;

		//'#'̳Ÿ  (NULLó)
		for (i = 0; i < strlen(key); i++) {
			if (key[i] == '"') {
				flag = !flag;
			}
			if (key[i] == '#' && flag == FALSE) {
				key[i] = '\0';
				break;
			}
		}

		//"  " ̹Ƿ  ߰ tap̳ ̽ ã 
		//  κ data  
		for (i = 0; i < strlen(key); i++) {
			if (key[i] == ' ' || key[i] == '\t') {
				key[i] = '\0';
				data = key + i + 1;
				break;
			}
		}
		if (strlen(key) != 0) { //key '#'  ʴ  
			result = (*process)(userdata, key, data);   
			if (result < 0) {
				fprintf(stderr,"load_data: key = %s, data = %s : ", key, data);
				fprintf(stderr, "result = %d\n", result);
				break;
			}
			if (result > 0) {
				fprintf(stderr,"load_data: key = %s, data = %s : ", key, data);
				fprintf(stderr, "result = %d\n", result);
				free(buf);
				return TRUE;
			}
		}
	} while (1);
	free(buf);
	return FALSE;
}

/* Load integer
 *
 * Integers starting with:
 * - "0x" are considered to be hexadecimal.
 * - "0" are considered to be octal.
 * - anything else is considered to be decimal.
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) x must not be NULL
 * M08) Hexadecimal number
 * M09) Octal number
 * M10) Decimal number
 * M11) Something else
 */

int load_int(const char *data, int *x)
{
	const char *pos = data;

	ASSERT(data != NULL);
	ASSERT(x != NULL);

	*x = 0;

	while (*pos == ' ' || *pos == '\t')
		pos++;

	if (strncmp(pos, "0x", 2) == 0) {
		if (sscanf(pos, "%x", x) != 1) return FALSE;
	} else if (strncmp(pos, "0", 1) == 0) {
		if (sscanf(pos, "%o", x) != 1) return FALSE;
	} else {
		if (sscanf(pos, "%d", x) != 1) return FALSE;
	}

	return TRUE;
}

/* Load boolean
 *
 * Boolean can be either "TRUE" or "FALSE"
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) x must not be NULL
 * M12) TRUE
 * M13) FALSE
 * M14) Something else
 */

int load_bool(const char *data, int *x)
{
	const char *pos = data;

	ASSERT(data != NULL);
	ASSERT(x != NULL);
	*x = FALSE;

	while (*pos == ' ' || *pos == '\t')
		pos++;

	if (strcmp(pos, "TRUE") == 0) {
		*x = TRUE;
		return TRUE;
	}
	if (strcmp(pos, "FALSE") == 0) {
		return TRUE;
	}
	return FALSE;
}


//XXXX:XXXX::XXX/dd ̾ Ѵ 
int load_ip6_addr(const char *data, struct in6_addr *addr, int * prefix_len)
{
	char tmp[100] = {0x00,}; 
	const char *pos;
	char *p, *n;

	pos = data;
	while (*pos == ' ' || *pos == '\t')
	pos++;

	if (sscanf(pos, "%s", tmp) != 1) {
		printf("load_ip6_addr: Unable to parse IP byte "
		"from string '%s'\n", pos);
		return FALSE;
	}
	memset(addr, 0, sizeof(struct in6_addr));
	* prefix_len =0;

	n = strchr(tmp, '/');
	if (n) {
		p = n + 1;
		*n = '\0';
		*prefix_len = strtoul(p, &p, 10);
		if (*p) 
			*prefix_len = 128;
	} else {
		printf("Invalid ip address: no prefix length\n");
		return FALSE;
	}
	if (inet_pton(AF_INET6, tmp, addr) <= 0) {
		printf("Invalid ip address: parse error\n");
		return FALSE;
	}
	printf("Load ipv6 address[%s]\n", tmp);
	return TRUE;
}


int get_hex_digit(const char c)
{
        if (c >= '0' && c <= '9') return (c - '0');
        if (c >= 'a' && c <= 'f') return (c - 'a' + 10);
        if (c >= 'A' && c <= 'F') return (c - 'A' + 10);
        return -1;
}


int load_link_addr(const char *data, unsigned char *addr )
{
	unsigned char tmp[30] = {0x00,}; 
	int i;
        const char *pos;
        char *n;
	
	pos = data;
	while (*pos == ' ' || *pos == '\t')
		pos++;

	if (sscanf(pos, "%s", tmp) != 1) {
		fprintf(stderr,
			"load_ip6_addr: Unable to parse MAC byte "
			"from string '%s'\n", pos);
		return FALSE;
	}

	i = strlen(tmp);
	tmp[i] = ':';

	pos = tmp;
	for(i = 0; i < 6 ;i ++){
		n = strchr(pos, ':');

		if (n) {
			*n = '\0';
			addr[i] = strtoul(pos, NULL, 16);
			pos =  n + 1;
		} else {
			if(i<5)
				return FALSE;
		}
	}
	return TRUE;
}

/* Load character table (string)
 *
 * NOTE! The routine loads actually a table of characters.
 * If the string is as long as the buffer, the string will not be
 * null-terminated! The user of the routine has to make sure that the
 * string is null-terminated if required.
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) table must not be NULL
 * M30) Valid string
 * M31) 2 x '"' in string
 * M32) only one '"' in string
 * M33) '"' in the middle string
 * M34) '\n' in string
 * M35) '\r' in string
 * M36) '\\' in string
 * M37) '\"' in string
 * M38) '\oct' in string
 * M39) Too long string
 */

int load_char_table(const char *data, char *table, int maxlen)
{
	int i, j, startpos;
	char temp[5];
	int val, write, len;
	char ch;
	int start, stop;

	ASSERT(data != NULL);
	ASSERT(table != NULL);
	for (i = 0 ; i < maxlen ; i++) {
		table[i] = '\0';
	}
	write = 0;
	start = FALSE;
	stop = FALSE;
	startpos = 0;
	while (data[startpos] == ' ' || data[startpos] == '\t')
		startpos++;
	len = strlen(data) - startpos;
	for (i = startpos; i <= len; i++) {
		val = data[i];
		if (val == '\0') {
			if (start == TRUE && stop == TRUE) break;
			fprintf(stderr,
				"load_char_table: Unexpected end of string\n");
			return FALSE;
		}
		if (val == '"') {
			if (start == FALSE) {
				start = TRUE;
				continue;
			} else if (stop == FALSE) {
				stop = TRUE;
				continue;
			} else {
				fprintf(stderr,
					"load_char_table: Unexpected '\"'\n");
				return FALSE;
			}
		} else {
			if (start == FALSE) {
				fprintf(stderr,
					"load_char_table: Unrecognized "
					"character '%c'\n",
					val);
				return FALSE;
			}
		}
		if (val == '\\') {
			if (i == len - 1) {
				fprintf(stderr,
					"load_char_table: String longer "
					"than expected\n");
				return FALSE;
			}
			i++;
			if (data[i] == '0') {
				ch = '\0';
			} else if (data[i] == 'n') {
				ch = '\n';
			} else if (data[i] == 'r') {
				ch = '\r';
			} else if (data[i] == '\\') {
				ch = '\\';
			} else if (data[i] == '"') {
				ch = '"';
			} else {
				if (i >= len - 3) {
					fprintf(stderr,
						"load_char_table: String "
						"longer than expected\n");
					return FALSE;
				}
				for (j = 0 ; j < 3 ; j++) {
					ch = data[i];
					i++;
					if (ch < '0' || ch > '9') {
						fprintf(stderr,
							"load_char_table: "
							"Invalid octal "
							"number\n");
						return FALSE;
					}
					temp[j] = ch;
				}
				i--;
				temp[3] = '\0';
				if (sscanf(temp, "%03o", &val) != 1) {
					fprintf(stderr,
						"load_char_table: Invalid "
						"octal number\n");
					return FALSE;
				}
				ch = val;
			}
		} else {
			ch = val;
		}
		if (write == maxlen) {
			fprintf(stderr,
				"load_char_table: String longer than "
				"expected\n");
			return FALSE;
		}
		table[write] = ch;
		write++;
	}
	if (write < maxlen) {
		table[write] = '\0';
	}
	return TRUE;
}

